/*
 * tag.c - Create an index from a tagged input file
 *         Due to the nature of the algorithm, anything that appears in
 *         the input file before the first tag is actually a comment
 *         since a key/offset value will not appear in the index for it.
 *
 * $Source: /u/pfratar/src/files-2.1/util/RCS/tag.c,v $
 *
 * $Log: tag.c,v $
 * Revision 1.2  1993/03/04  23:17:59  pfratar
 * Made a few minor formatting changes and added a few comments.
 *
 * Revision 1.1  1993/03/04  23:03:13  pfratar
 * Initial revision
 *
 */

#ifndef lint
static char *rcsid = "$Id: tag.c,v 1.2 1993/03/04 23:17:59 pfratar Exp $";
#endif

#include <stdio.h>
#include <string.h>
#include <strings.h>

#define KEY_TAG		"@@key:"
#define KEY_TAG_LEN	6
#define TABLE_SIZE	256

char Pname[1024];

/*
 * Table structure to hold the key and the offset for that key
 */
typedef struct tab {
    int	ind;
    long	off;
} tab_t;

/*
 * The user fucked up, print out a usage message to help them
 */
void
usage(void)
{
    printf("Usage: %s [ infile [ outfile ] ]\n", Pname);
    return;
}

/*
 * We want to sort the table of key/offset pairs, here is our comparison
 * routine to pass to qsort
 */
int
compar(tab_t *a, tab_t *b)
{
    if(a->ind > b->ind) return(1);
    else if(a->ind < b->ind) return(-1);
    else return(0);
}

/*
 * Run through the input file, find the keys, grab the offset, store
 * the key/value pairs in the table, sort the table, output the
 * table to the output file
 */
int
create_index(FILE *fp, FILE *ofp)
{
    char	line[1024], *p, *q;
    long	current;
    tab_t	*table, *t;
    int	tsize, tind = 0, i;

    /*
     * Get an initial table
     */
    tsize = TABLE_SIZE;
    table = (tab_t *)malloc(tsize * sizeof(tab_t));
    if(!table) return(1);
    t = table;

    while(fgets(line, sizeof line, fp)) {

        /* Get the current offset in the file */
        current = ftell(fp);

        /* Did we just read a key tag? */
        if(!strncmp(line, KEY_TAG, KEY_TAG_LEN)) {

            /*
             * We have found a new key, do we have enough space
             * in the table?  If not, allocate more space.
             */
            if(tind >= tsize) {
                tsize *= 2;
                table = (tab_t *)realloc(table, tsize * sizeof(tab_t));
                t = table + tind;
            }

            /* NULL out the newline */
            line[strlen(line) - 1] = '\0';

            /* Skip over the tag */
            p = line + KEY_TAG_LEN;

            /* Skip over white space */
            while(*p && *p == ' ') p++;

            /*
             * Convert the comma separated numbers and place
             * into the table.
             */
            q = p;
            while(*q) {
                /* Find a comma or EOL */
                while(*p && *p != ',') p++;
                if(*p == ',') {
                    *p = '\0';
                    t->ind = atoi(q);
                    t->off = current;
                    *p++ = ',';
                } else {
                    t->ind = atoi(q);
                    t->off = current;
                }
                t++, tind++;
                q = p;
            }
        }
    }

    /* Sort the index list */
    qsort(table, tind, sizeof(tab_t), compar);

    /* Print out the index list */
    t = table;
    for(i = 0; i < tind; i++) {
        sprintf(line, "%d:%ld\n", t->ind, t->off);
        fputs(line, ofp);
        t++;
    }

    if(table) free(table);
    return(0);
}

int
main(int argc, char **argv)
{
    char	*cp;
    FILE	*ifp, *ofp;

    /*
     * Get the program name
     */
    if(cp = rindex(argv[0], '/'))
        strcpy(Pname, cp+1);
    else
        strcpy(Pname, argv[0]);

    /*
     * Assume initially that we will be using stdin and stdout.
     */
    ifp = stdin;
    ofp = stdout;

    /*
     * If there were arguments, do create_index() once for each,
     * otherwise, use stdin.
     */
    switch(argc) {
    case 1:
        if(create_index(ifp, ofp))
            fprintf(stderr, "%s: %s\n", Pname, "Error creating index");
        break;
    case 3:
        if((ofp = fopen(argv[2], "w")) == NULL) {
            fprintf(stderr, "%s: %s '%s'\n", Pname,
                "Could not open output file", argv[2]
            );
        }
        /* FALL THROUGH */
    case 2:
        if((ifp = fopen(argv[1], "r")) == NULL) {
            fprintf(stderr, "%s: %s '%s'\n",
                Pname, "Could not open file", argv[1]
            );
        } else {
            if(create_index(ifp, ofp))
                fprintf(stderr, "%s: %s '%s'\n",
                    Pname, "Error creating index for", argv[1]
                );
            fclose(ifp);
            fclose(ofp);
        }
        break;
    default:
        fprintf(stderr, "%s: Invalid arguments\n", Pname);
        usage();
        break;
    }
    return(0);
}

